#include <stdio.h>
#include <string.h>
#include <spurs/spsa/spsa.h>
#include <spu_mfcio.h>
#include <spu_intrinsics.h>

#define DECREMENTER_INIT 0xffffffff

char buffer[0x20000] __attribute__((aligned(128)));


// vector単位でコピー(SIMD、ループアンローリングで最適化)
void cpy_optimise(void * dst, void * src, unsigned int size)
{
  vector unsigned int * vdst = dst;
  vector unsigned int * vsrc = src;
  unsigned int sizev = size >> 4;
  unsigned int to = sizev / 7;
  unsigned int i;
  unsigned int j = 0;
  for (i = 0; i < to; i++)
  {
    register vector unsigned int temp[7];
    temp[0] = vsrc[j + 0];
    temp[1] = vsrc[j + 1];
    temp[2] = vsrc[j + 2];
    temp[3] = vsrc[j + 3];
    temp[4] = vsrc[j + 4];
    temp[5] = vsrc[j + 5];
    temp[6] = vsrc[j + 6];
    vdst[j + 0] = temp[0];
    vdst[j + 1] = temp[1];
    vdst[j + 2] = temp[2];
    vdst[j + 3] = temp[3];
    vdst[j + 4] = temp[4];
    vdst[j + 5] = temp[5];
    vdst[j + 6] = temp[6];
    j += 7;
  }

  // 半端な残りを処理
  for (i = j; i < sizev; i++)
  {
    vdst[i] = vsrc[i];
  }
}


// vector単位で単純にコピー
void cpy_vec(void * dst, void * src, unsigned int size)
{
  vector unsigned int * vdst = dst;
  vector unsigned int * vsrc = src;
  unsigned int sizev = size >> 4;
  unsigned int i;
  for (i = 0; i < sizev; i++)
  {
    vdst[i] = vsrc[i];
  }
}


// memcpyでコピー
void cpy_memcpy(void * dst, void * src, unsigned int size)
{
  memcpy(dst, src, size);
}


// 1バイトずつ単純にコピー
void cpy_byte(void * dst, void * src, unsigned int size)
{
  char * cdst = dst;
  char * csrc = src;
  unsigned int i;
  for (i = 0; i < size; i++)
  {
    cdst[i] = csrc[i];
  }
}


int main(vector unsigned int arg0, vector unsigned int arg1, vector unsigned int arg2)
{
  unsigned int start, end;

  // SPU Decrementerに初期値をセット
  spu_writech(SPU_WrDec, DECREMENTER_INIT);

  // 計測開始
  start = spu_readch(SPU_RdDec);

  // 実行
  cpy_optimise(buffer + 0x10000, buffer, 0x10000);

  // 計測終了
  end = spu_readch(SPU_RdDec);

  // 結果表示
  printf("optimized LS memory copy: %f GB/s\n", (float)0x10000 / (float)(1024 * 1024 * 1024) / ((float)(start - end) / 16666666.7f));

  // 参考：vector単位で単純にコピーした場合
  spu_writech(SPU_WrDec, DECREMENTER_INIT);
  start = spu_readch(SPU_RdDec);
  cpy_vec(buffer + 0x10000, buffer, 0x10000);
  end = spu_readch(SPU_RdDec);
  printf("simple vector copy: %f GB/s\n", (float)0x10000 / (float)(1024 * 1024 * 1024) / ((float)(start - end) / 16666666.7f));

  // 参考：memcpyを使った場合
  spu_writech(SPU_WrDec, DECREMENTER_INIT);
  start = spu_readch(SPU_RdDec);
  cpy_memcpy(buffer + 0x10000, buffer, 0x10000);
  end = spu_readch(SPU_RdDec);
  printf("memcpy: %f GB/s\n", (float)0x10000 / (float)(1024 * 1024 * 1024) / ((float)(start - end) / 16666666.7f));

  // 参考：1Byteずつ単純にコピーした場合
  spu_writech(SPU_WrDec, DECREMENTER_INIT);
  start = spu_readch(SPU_RdDec);
  cpy_byte(buffer + 0x10000, buffer, 0x10000);
  end = spu_readch(SPU_RdDec);
  printf("simple byte copy: %f GB/s\n", (float)0x10000 / (float)(1024 * 1024 * 1024) / ((float)(start - end) / 16666666.7f));

  return 0;
}
